@@ -2,6 +2,7 @@
 
 from __future__ import annotations
 
+import bisect
 import typing
 import warnings
 
@@ -219,6 +220,12 @@ class Envelope(
     #                         private methods                                #
     # ###################################################################### #
 
+    def _make_event(self, duration, parameter, curve_shape):
+        event = self.initialise_default_event_class(self.default_event_class, duration)
+        self.apply_parameter_on_event(event, parameter)
+        self.apply_curve_shape_on_event(event, curve_shape)
+        return event
+
     def _point_sequence_to_event_list(
         self,
         point_or_invalid_type_sequence: typing.Sequence[
@@ -241,11 +248,7 @@ class Envelope(
             else:
                 absolute_time1 = absolute_time0
             duration = absolute_time1 - absolute_time0
-            event = self.initialise_default_event_class(
-                self.default_event_class, duration
-            )
-            self.apply_parameter_on_event(event, value_or_parameter)
-            self.apply_curve_shape_on_event(event, curve_shape)
+            event = self._make_event(duration, value_or_parameter, curve_shape)
             event_list.append(event)
         return event_list
 
@@ -307,13 +310,23 @@ class Envelope(
     def value_at(
         self, absolute_time: typing.Union[core_parameters.abc.Duration, typing.Any]
     ) -> Value:
-        absolute_time = core_events.configurations.UNKNOWN_OBJECT_TO_DURATION(
+        absolute_time_in_floats = core_events.configurations.UNKNOWN_OBJECT_TO_DURATION(
             absolute_time
-        )
-        absolute_time_tuple = self.absolute_time_tuple
+        ).duration_in_floats
 
-        use_only_first_event = absolute_time <= absolute_time_tuple[0]
-        use_only_last_event = absolute_time >= (
+        (
+            absolute_time_in_floats_tuple,
+            duration_in_floats,
+        ) = self._absolute_time_in_floats_tuple_and_duration
+
+        try:
+            use_only_first_event = (
+                absolute_time_in_floats <= absolute_time_in_floats_tuple[0]
+            )
+        except IndexError:
+            raise core_utilities.EmptyEnvelopeError(self, "value_at")
+
+        use_only_last_event = absolute_time_in_floats >= (
             # If the duration of the last event == 0 there is the danger
             # of floating point errors (the value in absolute_time_tuple could
             # be slightly higher than the duration of the Envelope. If this
@@ -321,16 +334,16 @@ class Envelope(
             # "_get_index_at_from_absolute_time_tuple" will return
             # "None"). With explicitly testing if the last duration
             # equals 0 we can avoid this danger.
-            absolute_time_tuple[-1]
+            absolute_time_in_floats_tuple[-1]
             if self[-1].duration > 0
-            else self.duration
+            else duration_in_floats
         )
         if use_only_first_event or use_only_last_event:
             index = 0 if use_only_first_event else -1
             return self._event_to_value(self[index])
 
         event_0_index = self._get_index_at_from_absolute_time_tuple(
-            absolute_time, absolute_time_tuple, self.duration
+            absolute_time, absolute_time_in_floats_tuple, duration_in_floats
         )
         assert event_0_index is not None
 
@@ -340,19 +353,110 @@ class Envelope(
         curve_shape = self.event_to_curve_shape(self[event_0_index])
 
         return core_utilities.scale(
-            absolute_time.duration_in_floats,
-            absolute_time_tuple[event_0_index].duration_in_floats,
-            absolute_time_tuple[event_0_index + 1].duration_in_floats,
+            absolute_time_in_floats,
+            absolute_time_in_floats_tuple[event_0_index],
+            absolute_time_in_floats_tuple[event_0_index + 1],
             value0,
             value1,
             curve_shape,
         )
 
     def parameter_at(
-        self, absolute_time: core_constants.DurationType
+        self, absolute_time: typing.Union[core_parameters.abc.Duration, typing.Any]
     ) -> core_constants.ParameterType:
         return self.value_to_parameter(self.value_at(absolute_time))
 
+    @core_utilities.add_copy_option
+    def sample_at(
+        self,
+        absolute_time: typing.Union[core_parameters.abc.Duration, typing.Any],
+        append_duration: typing.Union[
+            core_parameters.abc.Duration, typing.Any
+        ] = core_parameters.DirectDuration(0),
+    ) -> Envelope:
+        """Discretize envelope at given time
+
+        :param absolute_time: Position in time where the envelope should
+            define a new event.
+        :type absolute_time: typing.Union[core_parameters.abc.Duration, typing.Any]
+        :param append_duration: In case we add a new control point after any
+            already defined point, the duration of this control point will be
+            equal to "append_duration". Default to core_parameters.DirectDuration(0)
+        """
+
+        def find_duration(
+            absolute_time: core_parameters.abc.Duration,
+            absolute_time_tuple: tuple[core_parameters.abc.Duration, ...],
+        ):
+            """Find duration of new control point"""
+            next_event_start_index = bisect.bisect_right(
+                absolute_time_tuple, absolute_time
+            )
+            try:
+                next_event_start = absolute_time_tuple[next_event_start_index]
+            # In case we call "sample_at" at a position after any already
+            # specified point.
+            except IndexError:
+                duration_new_event = append_duration
+            else:
+                duration_new_event = next_event_start - absolute_time
+
+            return duration_new_event
+
+        def find_curve_shape(
+            absolute_time: core_parameters.abc.Duration,
+            absolute_time_tuple: tuple[core_parameters.abc.Duration, ...],
+            envelope_duration: core_parameters.abc.Duration,
+        ):
+            """Find curve shape of new control point"""
+            old_event_index = (
+                core_events.SequentialEvent._get_index_at_from_absolute_time_tuple(
+                    absolute_time, absolute_time_tuple, envelope_duration
+                )
+            )
+            if old_event_index is not None:
+                old_event = self[old_event_index]
+                curve_shape = self.event_to_curve_shape(old_event)
+                curve_shape_old_event = (
+                    (absolute_time - absolute_time_tuple[old_event_index])
+                    / old_event.duration
+                ).duration_in_floats * curve_shape
+                curve_shape_new_event = curve_shape - curve_shape_old_event
+                self.apply_curve_shape_on_event(old_event, curve_shape_old_event)
+            else:
+                curve_shape_new_event = 0
+
+            return curve_shape_new_event
+
+        if not self:
+            raise core_utilities.EmptyEnvelopeError(self, "sample_at")
+
+        absolute_time, append_duration = (
+            core_events.configurations.UNKNOWN_OBJECT_TO_DURATION(unknown_object)
+            for unknown_object in (absolute_time, append_duration)
+        )
+
+        # We only add a new event in case there isn't any event yet at
+        # given point in time.
+        if absolute_time not in (absolute_time_tuple := self.absolute_time_tuple):
+            envelope_duration = absolute_time_tuple[-1] + self[-1].duration
+            event = self._make_event(
+                find_duration(absolute_time, absolute_time_tuple),
+                self.parameter_at(absolute_time),
+                find_curve_shape(absolute_time, absolute_time_tuple, envelope_duration),
+            )
+
+            try:
+                self.squash_in(absolute_time, event)
+            # This means we want to squash in at a position much
+            # later than any already defined event.
+            except core_utilities.InvalidStartValueError:
+                difference = absolute_time - envelope_duration
+                self[-1].duration += difference
+                self.append(event)
+
+        return self
+
     def integrate_interval(
         self, start: core_constants.DurationType, end: core_constants.DurationType
     ) -> float:
@@ -390,6 +494,66 @@ class Envelope(
     ) -> core_constants.ParameterType:
         return self.value_to_parameter(self.get_average_value(start, end))
 
+    @core_utilities.add_copy_option
+    def cut_out(
+        self,
+        start: typing.Union[core_parameters.abc.Duration, typing.Any],
+        end: typing.Union[core_parameters.abc.Duration, typing.Any],
+    ) -> Envelope[T]:
+        start, end = (
+            core_events.configurations.UNKNOWN_OBJECT_TO_DURATION(unknown_object)
+            for unknown_object in (start, end)
+        )
+
+        self.sample_at(start, append_duration=end - start)
+        self.sample_at(end)
+
+        last_point = self.get_event_at(end)
+
+        # In case last_point.duration == 0 "get_event_at" won't return
+        # any object. This only happens in case
+        #
+        #   end > self.duration
+        #
+        # So the new point will be appended.
+        if last_point is None:
+            last_point = self[-1]
+        assert last_point
+
+        cut_out_envelope = super().cut_out(start, end)
+        cut_out_envelope.append(last_point.set("duration", 0))
+
+        return cut_out_envelope
+
+    @core_utilities.add_copy_option
+    def cut_off(
+        self,
+        start: typing.Union[core_parameters.abc.Duration, typing.Any],
+        end: typing.Union[core_parameters.abc.Duration, typing.Any],
+    ) -> Envelope[T]:
+        start, end = (
+            core_events.configurations.UNKNOWN_OBJECT_TO_DURATION(unknown_object)
+            for unknown_object in (start, end)
+        )
+
+        if (cut_off_duration := end - start) > 0:
+            # XXX: It is sufficient to find the first control point
+            # by simply using "parameter_at" instead of "sample_at":
+            # We don't need an accurate curve_shape or duration,
+            # because this point only exists in an infinitely short
+            # moment in time anyway (or better: its main function is
+            # to ensure that interpolation from the previous point
+            # to this point works as expected).
+            parameter_0 = self.parameter_at(start)
+            event_0 = self._make_event(0, parameter_0, 0)
+
+            self.sample_at(end)
+
+            self._cut_off(start, end, cut_off_duration)
+            self.squash_in(start, event_0)
+
+        return self
+
 
 class RelativeEnvelope(Envelope, typing.Generic[T]):
     __parent_doc_string = Envelope.__doc__.split("\n")[2:]  # type: ignore
@@ -452,7 +616,101 @@ class RelativeEnvelope(Envelope, typing.Generic[T]):
         return resolve_envelope_class(point_list)
 
 
+TempoPoint = typing.Union[core_parameters.abc.TempoPoint, core_constants.Real]
+
+
 class TempoEnvelope(Envelope):
+    """Define dynamic or static tempo trajectories.
+
+    You can either define a new `TempoEnvelope` with instances
+    of classes which inherit from :class:`mutwo.core_parameters.abc.TempoPoint`
+    (for instance :class:`mutwo.core_parameters.DirectTempoPoint`) or with
+    `float` or `int` objects which represent beats per minute.
+
+    Please see the :class:`mutwo.core_events.Envelope` for full documentation
+    for initialization attributes.
+
+    The default parameters of the `TempoEnvelope` class expects
+    :class:`mutwo.core_events.SimpleEvent` to which a tempo point
+    was assigned by the name "tempo_point". This is specified in the global
+    `mutwo.core_events.configurations.DEFAULT_TEMPO_ENVELOPE_PARAMETER_NAME`
+    and can be adjusted.
+
+    **Example:**
+
+    >>> from mutwo import core_events
+    >>> from mutwo import core_parameters
+    >>> # (1) define with floats
+    >>> #     So we have an envelope which moves from tempo 60 to 30
+    >>> #     and back to 60.
+    >>> tempo_envelope_with_float = core_events.TempoEnvelope(
+    >>>     [[0, 60], [1, 30], [2, 60]]
+    >>> )
+    >>> # (2) define with tempo points
+    >>> tempo_envelope_with_tempo_points = core_events.TempoEnvelope(
+    >>>     [
+    >>>         [0, core_parameters.DirectTempoPoint(60)],
+    >>>         [1, core_parameters.DirectTempoPoint(30)],
+    >>>         [2, core_parameters.DirectTempoPoint(30, reference=2)],
+    >>>     ]
+    >>> )
+    """
+
+    def __init__(
+        self,
+        *args,
+        event_to_parameter: typing.Optional[
+            typing.Callable[[core_events.abc.Event], core_constants.ParameterType]
+        ] = None,
+        value_to_parameter: typing.Optional[
+            typing.Callable[[core_events.Envelope.Value], core_constants.ParameterType]
+        ] = None,
+        parameter_to_value: typing.Optional[
+            typing.Callable[[core_constants.ParameterType], core_events.Envelope.Value]
+        ] = None,
+        apply_parameter_on_event: typing.Optional[
+            typing.Callable[[core_events.abc.Event, core_constants.ParameterType], None]
+        ] = None,
+        **kwargs,
+    ):
+        def default_event_to_parameter(event: core_events.abc.Event) -> TempoPoint:
+            return getattr(
+                event,
+                core_events.configurations.DEFAULT_TEMPO_ENVELOPE_PARAMETER_NAME,
+            )
+
+        def default_value_to_parameter(value: float) -> TempoPoint:
+            return core_parameters.DirectTempoPoint(value)
+
+        def default_parameter_to_value(parameter: TempoPoint) -> float:
+            # XXX: Here we specify, that we allow either core_parameters.abc.TempoPoint
+            # or float/number objects.
+            # So in case we have a core_parameters.abc.TempoPoint 'getattr' is
+            # successful, if not it will return 'parameter', because it
+            # will assume that we have a number based tempo point.
+            return float(
+                getattr(parameter, "absolute_tempo_in_beats_per_minute", parameter)
+            )
+
+        def default_apply_parameter_on_event(
+            event: core_events.abc.Event, parameter: TempoPoint
+        ):
+            setattr(
+                event,
+                core_events.configurations.DEFAULT_TEMPO_ENVELOPE_PARAMETER_NAME,
+                parameter,
+            )
+
+        super().__init__(
+            *args,
+            event_to_parameter=event_to_parameter or default_event_to_parameter,
+            value_to_parameter=value_to_parameter or default_value_to_parameter,
+            parameter_to_value=parameter_to_value or default_parameter_to_value,
+            apply_parameter_on_event=apply_parameter_on_event
+            or default_apply_parameter_on_event,
+            **kwargs,
+        )
+
     def __eq__(self, other: typing.Any):
         # XXX: TempoEnvelope can't use the default '__eq__' method inherited
         # from list, because this would create endless recursion
